#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2012 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Operators;
using HeuristicLab.Optimization;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Encodings.ConditionActionEncoding {
  [Item("XCSDeletionOperator", "Description missing")]
  [StorableClass]
  public class XCSDeletionOperator : SingleSuccessorOperator, IStochasticOperator {

    #region Parameter Properties
    public ILookupParameter<IntValue> NumberToDeleteParameter {
      get { return (ILookupParameter<IntValue>)Parameters["NumberToDelete"]; }
    }
    public ILookupParameter<ItemArray<IntValue>> NumerositiesParameter {
      get { return (ILookupParameter<ItemArray<IntValue>>)Parameters["Numerosities"]; }
    }
    public ILookupParameter<ItemArray<DoubleValue>> AverageActionSetSizesParameter {
      get { return (ILookupParameter<ItemArray<DoubleValue>>)Parameters["AverageActionSetSizes"]; }
    }
    public ILookupParameter<ItemArray<DoubleValue>> FitnessesParameter {
      get { return (ILookupParameter<ItemArray<DoubleValue>>)Parameters["Fitnesses"]; }
    }
    public ILookupParameter<ItemArray<IntValue>> ExperiencesParameter {
      get { return (ILookupParameter<ItemArray<IntValue>>)Parameters["Experiences"]; }
    }
    public ILookupParameter<ItemArray<BoolValue>> HasToBeDeletedParameter {
      get { return (ILookupParameter<ItemArray<BoolValue>>)Parameters["HasToBeDeleted"]; }
    }
    public ILookupParameter<IntValue> ThetaDeletionParameter {
      get { return (ILookupParameter<IntValue>)Parameters["ThetaDeletion"]; }
    }
    public ILookupParameter<PercentValue> DeltaParameter {
      get { return (ILookupParameter<PercentValue>)Parameters["Delta"]; }
    }
    public ILookupParameter<IRandom> RandomParameter {
      get { return (ILookupParameter<IRandom>)Parameters["Random"]; }
    }
    public ILookupParameter<IntValue> CurrentPopulationSizeParameter {
      get { return (ILookupParameter<IntValue>)Parameters["CurrentPopulationSize"]; }
    }
    #endregion

    #region Parameter Properties
    private double Delta { get { return DeltaParameter.ActualValue.Value; } }
    private int ThetaDeletion { get { return ThetaDeletionParameter.ActualValue.Value; } }
    #endregion

    [StorableConstructor]
    protected XCSDeletionOperator(bool deserializing) : base(deserializing) { }
    protected XCSDeletionOperator(XCSDeletionOperator original, Cloner cloner)
      : base(original, cloner) {
    }
    public XCSDeletionOperator()
      : base() {
      Parameters.Add(new LookupParameter<IntValue>("NumberToDelete"));
      Parameters.Add(new ScopeTreeLookupParameter<IntValue>("Numerosities"));
      Parameters.Add(new ScopeTreeLookupParameter<DoubleValue>("AverageActionSetSizes"));
      Parameters.Add(new ScopeTreeLookupParameter<DoubleValue>("Fitnesses"));
      Parameters.Add(new ScopeTreeLookupParameter<IntValue>("Experiences"));
      Parameters.Add(new ScopeTreeLookupParameter<BoolValue>("HasToBeDeleted"));
      Parameters.Add(new LookupParameter<IntValue>("ThetaDeletion"));
      Parameters.Add(new LookupParameter<PercentValue>("Delta"));
      Parameters.Add(new LookupParameter<IRandom>("Random"));
      Parameters.Add(new LookupParameter<IntValue>("CurrentPopulationSize"));
    }
    public override IDeepCloneable Clone(Cloner cloner) {
      return new XCSDeletionOperator(this, cloner);
    }

    public override IOperation Apply() {
      var fitnesses = FitnessesParameter.ActualValue;
      var numerosities = NumerositiesParameter.ActualValue;
      var experiences = ExperiencesParameter.ActualValue;
      var averageActionSetSizes = AverageActionSetSizesParameter.ActualValue;
      var hasToBeDeleted = HasToBeDeletedParameter.ActualValue;
      double fitnessSum = fitnesses.Select(x => x.Value).Sum();
      double numerositySum = numerosities.Select(x => x.Value).Sum();

      if (fitnesses.Length != numerosities.Length && fitnesses.Length != experiences.Length) {
        throw new ArgumentException("Different number of fitness, numerosity and experience values.");
      }

      for (int i = 0; i < NumberToDeleteParameter.ActualValue.Value; i++) {
        double voteSum = 0;
        double averageFitnessInPopulation = fitnessSum / numerositySum;
        for (int cl = 0; cl < fitnesses.Length; cl++) {
          if (numerosities[cl].Value > 0) {
            voteSum += DeletionVote(averageFitnessInPopulation, averageActionSetSizes[cl].Value, numerosities[cl].Value, fitnesses[cl].Value, experiences[cl].Value);
          }
        }
        double choicePoint = RandomParameter.ActualValue.NextDouble() * voteSum;
        voteSum = 0;
        for (int cl = 0; cl < fitnesses.Length; cl++) {
          if (numerosities[cl].Value > 0 && !hasToBeDeleted[cl].Value) {
            voteSum += DeletionVote(averageFitnessInPopulation, averageActionSetSizes[cl].Value, numerosities[cl].Value, fitnesses[cl].Value, experiences[cl].Value);
            if (voteSum > choicePoint) {
              if (numerosities[cl].Value <= 1) {
                hasToBeDeleted[cl].Value = true;
                fitnessSum -= fitnesses[cl].Value;
              }
              numerosities[cl].Value--;
              numerositySum--;
              break;
            }
          }
        }
      }

      CurrentPopulationSizeParameter.ActualValue.Value -= NumberToDeleteParameter.ActualValue.Value;

      return base.Apply();
    }

    private double DeletionVote(double averageFitnessInPopulation, double ass, int num, double f, int exp) {
      double vote = ass * num;
      if (exp > ThetaDeletion && (f / num) < Delta * averageFitnessInPopulation) {
        vote *= averageFitnessInPopulation / (f / num);
      }
      return vote;
    }
  }
}