﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2012 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Operators;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Encodings.ConditionActionEncoding {
  [Item("CheckGASubsumptionOperator", "Description missing")]
  [StorableClass]
  public class CheckGASubsumptionOperator : SingleSuccessorOperator {

    #region Parameter Properties
    public ILookupParameter<IClassifier> ChildClassifiersParameter {
      get { return (ILookupParameter<IClassifier>)Parameters["ChildClassifiers"]; }
    }
    public ILookupParameter<ItemArray<IClassifier>> ParentsClassifiersParameter {
      get { return (ILookupParameter<ItemArray<IClassifier>>)Parameters["Classifiers"]; }
    }
    public ILookupParameter<ItemArray<IntValue>> NumerositiesParameter {
      get { return (ILookupParameter<ItemArray<IntValue>>)Parameters["Numerosities"]; }
    }
    public ILookupParameter<ItemArray<IntValue>> ExperiencesParameter {
      get { return (ILookupParameter<ItemArray<IntValue>>)Parameters["Experiences"]; }
    }
    public ILookupParameter<ItemArray<DoubleValue>> ErrorsParameter {
      get { return (ILookupParameter<ItemArray<DoubleValue>>)Parameters["Errors"]; }
    }
    public ILookupParameter<ItemArray<IntValue>> TempIDParameter {
      get { return (ILookupParameter<ItemArray<IntValue>>)Parameters["TempID"]; }
    }
    public ILookupParameter<DoubleValue> ErrorZeroParameter {
      get { return (ILookupParameter<DoubleValue>)Parameters["ErrorZero"]; }
    }
    public ILookupParameter<IntValue> ThetaSubsumptionParameter {
      get { return (ILookupParameter<IntValue>)Parameters["ThetaSubsumption"]; }
    }
    public IValueLookupParameter<IntValue> SubsumedByParameter {
      get { return (IValueLookupParameter<IntValue>)Parameters["SubsumedBy"]; }
    }
    public IValueLookupParameter<BoolValue> SubsumedParameter {
      get { return (IValueLookupParameter<BoolValue>)Parameters["Subsumed"]; }
    }
    #endregion

    [StorableConstructor]
    protected CheckGASubsumptionOperator(bool deserializing) : base(deserializing) { }
    protected CheckGASubsumptionOperator(CheckGASubsumptionOperator original, Cloner cloner)
      : base(original, cloner) {
    }
    public CheckGASubsumptionOperator()
      : base() {
      Parameters.Add(new LookupParameter<IClassifier>("ChildClassifiers"));
      Parameters.Add(new ScopeTreeLookupParameter<IClassifier>("Classifiers"));
      Parameters.Add(new ScopeTreeLookupParameter<IntValue>("Numerosities"));
      Parameters.Add(new ScopeTreeLookupParameter<IntValue>("Experiences"));
      Parameters.Add(new ScopeTreeLookupParameter<DoubleValue>("Errors"));
      Parameters.Add(new ScopeTreeLookupParameter<IntValue>("TempID"));
      Parameters.Add(new LookupParameter<DoubleValue>("ErrorZero"));
      Parameters.Add(new LookupParameter<IntValue>("ThetaSubsumption"));
      Parameters.Add(new ValueLookupParameter<IntValue>("SubsumedBy"));
      Parameters.Add(new ValueLookupParameter<BoolValue>("Subsumed"));
    }
    public override IDeepCloneable Clone(Cloner cloner) {
      return new CheckGASubsumptionOperator(this, cloner);
    }

    public override IOperation Apply() {
      var child = ChildClassifiersParameter.ActualValue;
      var parents = ParentsClassifiersParameter.ActualValue;
      var numerosities = NumerositiesParameter.ActualValue;
      var experiences = ExperiencesParameter.ActualValue;
      var tempIDs = TempIDParameter.ActualValue;
      var errors = ErrorsParameter.ActualValue;
      if (errors.Length != parents.Length || errors.Length != numerosities.Length
        || errors.Length != experiences.Length || errors.Length != tempIDs.Length) {
        throw new ArgumentException("The number of classifiers, error, numerosity, tempIDs and experience values is not equal.");
      }

      SubsumedByParameter.ActualValue = new IntValue(-1);
      SubsumedParameter.ActualValue = new BoolValue(false);

      for (int i = 0; i < parents.Length; i++) {
        IClassifier curParent = parents[i];
        if (curParent.MatchAction(child)) {
          if (CouldSubsume(experiences[i], errors[i])) {
            if (curParent.IsMoreGeneral(child)) {
              SubsumedByParameter.ActualValue.Value = i;
              SubsumedParameter.ActualValue.Value = true;
              break;
            }
          }
        }
      }

      return base.Apply();
    }

    private bool CouldSubsume(IntValue exp, DoubleValue error) {
      return exp.Value > ThetaSubsumptionParameter.ActualValue.Value && error.Value < ErrorZeroParameter.ActualValue.Value;
    }
  }
}
